'''
Graphics tests
==============

Testing the simple vertex instructions
'''

import sys
import pytest
import itertools
from threading import Thread
from kivy.tests.common import GraphicUnitTest, requires_graphics


class BoxShadowTest(GraphicUnitTest):

    def test_create(self):
        from kivy.graphics.boxshadow import BoxShadow
        from kivy.uix.widget import Widget
        from kivy.graphics import Color

        r = self.render

        # with initial arguments
        wid = Widget()
        with wid.canvas:
            Color(1, 0, 0, 1)
            bs = BoxShadow(
                pos=(50, 50),
                size=(150, 150),
                offset=(0, 10),
                spread_radius=(10, -10),
                border_radius=(10, 10, 10, 10),
                blur_radius=80,
            )
        r(wid)

        wid = Widget()
        with wid.canvas:
            Color(0, 1, 0, 1)
            bs = BoxShadow(
                inset=True,
                pos=(50, 50),
                size=(150, 150),
                offset=(0, 10),
                spread_radius=(10, -10),
                border_radius=(10, 10, 10, 10),
                blur_radius=80,
            )
        r(wid)

        # changing properties later
        wid = Widget()
        with wid.canvas:
            Color(0, 0, 1, 1)
            bs = BoxShadow()
        bs.inset = True
        bs.pos = [50, 50]
        bs.size = [150, 150]
        bs.offset = [0, 10]
        bs.spread_radius = [10, -10]
        bs.border_radius = [10, 10, 10, 10]
        bs.blur_radius = 40
        r(wid)

    def test_adjusted_size(self):
        from kivy.graphics.boxshadow import BoxShadow

        raw_size = 150, 150

        bs = BoxShadow()
        bs.pos = 50, 50
        bs.size = raw_size
        bs.blur_radius = 80
        bs.spread_radius = -10, 10

        # The size of the rectangle containing the FBO texture (shadow) needs
        # to be adjusted according to the size of the shadow, otherwise there
        # will be an unwanted cropping behavior.
        adjusted_size = (
            max(
                0,
                raw_size[0] + bs.blur_radius * 3 + bs.spread_radius[0] * 2,
            ),
            max(
                0,
                raw_size[1] + bs.blur_radius * 3 + bs.spread_radius[1] * 2,
            ),
        )

        assert bs.size == adjusted_size

        # Now we will turn on the inset mode, it is expected that
        # there will be no size adjustments.
        bs.inset = True
        assert bs.size == raw_size

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.size == adjusted_size

        # Testing with initial arguments
        bs = BoxShadow(
            inset=True,
            pos=(50, 50),
            size=raw_size,
            blur_radius=80,
            spread_radius=(10, -10)
        )
        adjusted_size = (
            max(
                0,
                raw_size[0] + bs.blur_radius * 3 + bs.spread_radius[0] * 2,
            ),
            max(
                0,
                raw_size[1] + bs.blur_radius * 3 + bs.spread_radius[1] * 2,
            ),
        )

        assert bs.size == raw_size

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.size == adjusted_size

        # Now we will turn on the inset mode, it is expected that
        # there will be no size adjustments.
        bs.inset = True
        assert bs.size == raw_size

    def test_adjusted_pos(self):
        from kivy.graphics.boxshadow import BoxShadow

        raw_pos = 50, 50
        raw_size = 150, 150
        offset = 10, -100

        bs = BoxShadow()
        bs.pos = raw_pos
        bs.size = raw_size
        bs.offset = offset
        bs.blur_radius = 80
        bs.spread_radius = -10, 10

        # If the size of the rectangle containing the FBO texture (shadow)
        # changes, its position will need to be adjusted.
        adjusted_pos = (
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[0]
            + bs.offset[0],
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[1]
            + bs.offset[1],
        )

        assert bs.pos == adjusted_pos

        # Now we will turn on the inset mode, it is expected that
        # there will be no position adjustments.
        bs.inset = True
        assert bs.pos == raw_pos

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.pos == adjusted_pos

        # Testing with initial arguments
        bs = BoxShadow(
            inset=True,
            pos=raw_pos,
            size=raw_size,
            offset=offset,
            blur_radius=80,
            spread_radius=(10, -10)
        )
        adjusted_pos = (
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[0]
            + bs.offset[0],
            raw_pos[0]
            - bs.blur_radius * 1.5
            - bs.spread_radius[1]
            + bs.offset[1],
        )

        assert bs.pos == raw_pos

        # Now turning off, and reverting back to the default mode.
        bs.inset = False
        assert bs.pos == adjusted_pos

        # Now we will turn on the inset mode, it is expected that
        # there will be no position adjustments.
        bs.inset = True
        assert bs.pos == raw_pos

    def test_bounded_properties(self):
        from kivy.graphics.boxshadow import BoxShadow

        bs = BoxShadow()
        bs.pos = 50, 50
        bs.size = 150, 150
        bs.offset = 10, -100
        bs.blur_radius = -80
        bs.spread_radius = -200, -100
        bs.border_radius = 0, 0, 100, 0

        assert bs.size == (0, 0)
        assert bs.blur_radius == 0

        # There is a bug in RoundedRectangle that distorts the texture if the
        # radius value is less than 1. Otherwise, it could be 0.
        assert bs.border_radius == tuple(
            map(
                lambda value: max(1.0, min(value, min(bs.size) / 2)),
                bs.border_radius,
            )
        )

        # Testing with initial arguments
        bs = BoxShadow(
            pos=(50, 50),
            size=(150, 150),
            offset=(10, -100),
            blur_radius=-80,
            spread_radius=(-200, -100),
            border_radius=(0, 0, 100, 0),
        )

        assert bs.size == (0, 0)
        assert bs.blur_radius == 0
        assert bs.border_radius == tuple(
            map(
                lambda value: max(1.0, min(value, min(bs.size) / 2)),
                bs.border_radius,
            )
        )

    def test_canvas_management(self):
        from kivy.graphics.boxshadow import BoxShadow
        from kivy.uix.widget import Widget
        from kivy.graphics import Color

        r = self.render

        wid = Widget()
        with wid.canvas:
            bs = BoxShadow()
        r(wid)
        assert bs in wid.canvas.children

        wid = Widget()
        bs = BoxShadow()
        wid.canvas.add(Color(1, 0, 0, 1))
        wid.canvas.add(bs)
        r(wid)
        assert bs in wid.canvas.children

        wid.canvas.remove(bs)
        assert bs not in wid.canvas.children

        wid.canvas.insert(1, bs)
        assert bs in wid.canvas.children
        assert wid.canvas.children.index(bs) == 1


class VertexInstructionTest(GraphicUnitTest):

    def test_circle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Ellipse, Color

        r = self.render

        # basic circle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Ellipse(pos=(100, 100), size=(100, 100))
        r(wid)

        # reduced circle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Ellipse(pos=(100, 100), size=(100, 100), segments=10)
        r(wid)

        # moving circle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            self.e = Ellipse(pos=(100, 100), size=(100, 100))
        self.e.pos = (10, 10)
        r(wid)

    def test_ellipse(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Ellipse, Color

        r = self.render

        # ellipse
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            self.e = Ellipse(pos=(100, 100), size=(200, 100))
        r(wid)

    def test_point(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Point, Color

        r = self.render

        # 1 point
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Point(points=(10, 10))
        r(wid)

        # 25 points
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            Point(points=[x * 5 for x in range(50)])
        r(wid)

    def test_point_add(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Point, Color

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            p = Point(pointsize=10)

        p.add_point(10, 10)
        p.add_point(90, 10)
        p.add_point(10, 90)
        p.add_point(50, 50)
        p.add_point(10, 50)
        p.add_point(50, 10)

        r(wid)

    def test_line_rounded_rectangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Line, Color
        r = self.render

        # basic rounded_rectangle
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(100, 100, 100, 100, 10, 20, 30, 40, 100)
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 10, 20, 30, 40, 100
        )

        # The largest angle allowed is equal to the smallest dimension (width
        # or height) minus the largest angle value between the anterior angle
        # and the posterior angle.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 100, 20, 10, 30, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 70, 20, 10, 30, 100
        )

        # Same approach as above
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 100, 25, 100, 50, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 50, 25, 50, 50, 100
        )

        # A circle should be generated if width and height are equal, and all
        # angles passed are greater than or equal to the smallest dimension.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 150, 50, 50.001, 51, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 50, 50, 50, 50, 100
        )

        # Currently the minimum radius should be 1, to avoid rendering issues
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 0, 0, 0, 0, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 1, 1, 1, 1, 100
        )

        # Angles adjustment + avoid issue if radius is less than 1
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = Line(
                rounded_rectangle=(
                    100, 100, 100, 100, 100, 0, 0, 0, 100
                )
            )
        r(wid)
        assert line.rounded_rectangle == (
            100, 100, 100, 100, 99, 1, 1, 1, 100
        )

    def test_smoothline_rounded_rectangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import SmoothLine, Color
        r = self.render

        # If width and/or height < 2px, the figure should not be rendered.
        # This avoids some known SmoothLine rendering issues.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)
            line = SmoothLine(
                rounded_rectangle=(100, 100, 0.5, 1.99, 30, 30, 30, 30, 100)
            )
        r(wid)
        assert line.rounded_rectangle is None

    def test_enlarged_line(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Line, Color, PushMatrix, PopMatrix, Scale, \
            Translate
        r = self.render
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1)

            # Normal line with width 1
            Line(
                points=(10, 10, 10, 90),
                width=1
            )

            # Normal line with width 3
            Line(
                points=(20, 10, 20, 90),
                width=3
            )

            # Enlarged line that should look width 3
            PushMatrix()
            Translate(30, 10, 1)  # So the enlargement goes around 0, 0, 0
            Scale(3, 1, 1)  # X scaled by 3 so the line width should become 3
            Line(
                points=(0, 0, 0, 80),
                width=1,
                force_custom_drawing_method=True
            )
            PopMatrix()

        r(wid)


class SmoothVertexInstructionTest(GraphicUnitTest):

    # Code defined in the points setter of AntiAliasingLine class.
    def _convert_points(self, points):
        if points and isinstance(points[0], (list, tuple)):
            return list(itertools.chain(*points))
        else:
            return list(points)

    # The same function present in the AntiAliasingLine code,
    # externalized for testing.
    def _filtered_points(self, points):
        index = 0
        p = self._convert_points(points)

        # At least 3 points are required, otherwise we will return an empty
        # list, which means there are no valid points.
        if len(p) < 6:
            return []

        while index < len(p) - 2:
            x1, y1 = p[index], p[index + 1]
            x2, y2 = p[index + 2], p[index + 3]
            if abs(x2 - x1) < 1.0 and abs(y2 - y1) < 1.0:
                del p[index + 2: index + 4]
            else:
                index += 2
        if abs(p[0] - p[-2]) < 1.0 and abs(p[1] - p[-1]) < 1.0:
            del p[:2]
        return p

    def _get_texture(self):
        from kivy.graphics.texture import Texture
        return Texture.create()

    def test_antialiasing_line(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Color, Rectangle, Instruction
        from kivy.graphics.vertex_instructions import AntiAliasingLine

        r = self.render

        # We expect a type error to be thrown if stencil_mask is not an object
        # of type Instruction.
        with pytest.raises(TypeError):
            AntiAliasingLine(None, points=[10, 20, 30, 20, 30, 10])

        # No TypeError here.
        target_rect = Rectangle()
        AntiAliasingLine(target_rect, points=[10, 20, 30, 40, 50, 60])

        # Test the gradient responsible for the "smooth line" effect.
        pixels = b"\xff\xff\xff\x00\xff\xff\xff\xff\xff\xff\xff\x00"
        instruction = Instruction()
        aa_line = AntiAliasingLine(instruction)
        assert aa_line.texture.pixels == pixels
        # check the width, defined through tests with the custom texture.
        assert aa_line.width == 2.5

        # This set of points must remain unchanged.
        points_1 = [51.0, 649.0, 199.0, 649.0, 199.0, 501.0, 51.0, 501.0]

        # This set of points should be reduced from 16 to 8.
        points_2 = [
            261.0, 275.0,
            335.0, 349.0,
            335.0, 349.0,
            409.0, 275.0,
            409.0, 275.0,
            335.0, 201.0,
            335.0, 201.0,
            261.0, 275.0
        ]

        # This set of points should be reduced from 72 to 50.
        points_3 = [
            260.0, 275.0,
            261.0, 275.0,
            261.0, 275.0,
            261.999999999999, 275.99999999,
            261.06667650085353, 278.14064903651496,
            261.26658584785304, 281.2756384111877,
            261.56658584785305, 281.3756384111877,
            261.5993677908431, 284.39931866126904,
            262.0644226342696, 287.50606070381684,
            262.0644226342696, 287.50606070381684,
            262.6609123178712, 290.59026597968375,
            263.3877619269211, 293.6463765424993,
            264.2436616292954, 296.66888507446475,
            265.22706903587977, 299.65234481091227,
            265.22706903587977, 299.65234481091227,
            266.3362119800583, 302.59137935574284,
            267.5690917112779, 305.48069237005546,
            268.9234864969319, 308.31507711650784,
            270.39695562607204, 311.089425842209,
            270.89695562607204, 311.589425842209,
            271.98684380773494, 313.7987389832352,
            273.69028595595563, 316.4381341741821,
            275.50421235284637, 319.00285504651725,
            275.50421235284637, 319.00285504651725,
            277.4253541804354, 321.48827979987755,
            279.45024941129833, 323.8899295308661,
            281.57524904736516, 326.20347630433844,
            283.79652369566156, 328.4247509526349,
            283.99652369566156, 328.7247509526349,
            286.1100704691339, 330.54975058870167,
            288.5117202001224, 332.5746458195646,
            288.5117202001224, 332.5746458195646,
            290.99714495348275, 334.4957876471537,
            293.5618658258179, 336.3097140440444,
            293.5618658258179, 336.3097140440444,
            293.2618658258179, 336.1097140440444
        ]

        # This set of points should be reduced to [].
        points_4 = [100, 100, 200, 100]

        # line closed (default)
        for points in (points_1, points_2, points_3, points_4):
            wid = Widget()
            with wid.canvas:
                Color(1, 1, 1, 0.5)
                inst = Instruction()
                aa_line = AntiAliasingLine(inst, points=points)
            r(wid)
            filtered_points = self._filtered_points(points)
            assert aa_line.points == filtered_points + filtered_points[:2]

        # without closing the line
        for points in (points_1, points_2, points_3, points_4):
            wid = Widget()
            with wid.canvas:
                Color(1, 1, 1, 0.5)
                inst = Instruction()
                aa_line = AntiAliasingLine(inst, points=points, close=0)
            r(wid)
            assert aa_line.points == self._filtered_points(points)

    def test_smoothrectangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Color, SmoothRectangle

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rect = SmoothRectangle(pos=(100, 100), size=(150, 150))
        r(wid)
        filtered_points = self._filtered_points(rect.points)
        assert (
            rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # test if antialiasing is disabled if the graphic has one of its
        # dimensions decreased to less than 4px.
        rect.size = (150, -2)
        r(wid)
        assert rect.antialiasing_line_points == []

        rect.size = (150, 2)
        r(wid)
        assert rect.antialiasing_line_points == []

        # re-enable antialiasing line rendering.
        rect.size = (150, 150)
        r(wid)
        assert (
            rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # disabling antialiasing.
        rect.texture = self._get_texture()
        r(wid)
        assert rect.antialiasing_line_points == []

        # re-enable antialiasing.
        rect.source = ""
        r(wid)
        assert (
            rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # test if antialiasing is disabled if the graphic is initialized with
        # at least one of its dimensions less than 4px.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rect = SmoothRectangle(pos=(100, 100), size=(150, -3))
        r(wid)
        assert rect.antialiasing_line_points == []

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rect = SmoothRectangle(pos=(100, 100), size=(3.99, 3.99))
        r(wid)
        assert rect.antialiasing_line_points == []

    def test_smoothroundedrectangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Color, SmoothRoundedRectangle

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rounded_rect = SmoothRoundedRectangle(
                pos=(100, 100),
                size=(150, 150),
                radius=[(10, 50), (100, 50), (0, 150), (200, 50)],
                segments=60,
            )
        r(wid)
        filtered_points = self._filtered_points(rounded_rect.points)
        assert (
            rounded_rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # test if antialiasing is disabled if the graphic has one of its
        # dimensions decreased to less than 4px.
        rounded_rect.size = (150, -2)
        r(wid)
        assert rounded_rect.antialiasing_line_points == []

        rounded_rect.size = (150, 2)
        r(wid)
        assert rounded_rect.antialiasing_line_points == []

        # re-enable antialiasing line rendering.
        rounded_rect.size = (150, 150)
        r(wid)
        assert (
            rounded_rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # disabling antialiasing.
        rounded_rect.texture = self._get_texture()
        r(wid)
        assert rounded_rect.antialiasing_line_points == []

        # re-enable antialiasing.
        rounded_rect.source = ""
        r(wid)
        assert (
            rounded_rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rounded_rect = SmoothRoundedRectangle(
                pos=(100, 100), size=(150, 150), segments=0
            )
        r(wid)
        filtered_points = self._filtered_points(rounded_rect.points)
        assert (
            rounded_rect.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # test if antialiasing is disabled if the graphic is initialized with
        # at least one of its dimensions less than 4px.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rounded_rect = SmoothRoundedRectangle(
                pos=(100, 100), size=(150, -3)
            )
        r(wid)
        assert rounded_rect.antialiasing_line_points == []

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            rounded_rect = SmoothRoundedRectangle(
                pos=(100, 100), size=(3.99, 3.99)
            )
        r(wid)
        assert rounded_rect.antialiasing_line_points == []

    def test_smoothellipse(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Color, SmoothEllipse

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            ellipse = SmoothEllipse(pos=(100, 100), size=(150, 150))
        r(wid)
        ellipse_center = [
            ellipse.pos[0] + ellipse.size[0] / 2,
            ellipse.pos[1] + ellipse.size[1] / 2,
        ]
        filtered_points = self._filtered_points(
            ellipse.points + ellipse_center
        )
        assert (
            ellipse.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # test if antialiasing is disabled if the graphic has one of its
        # dimensions decreased to less than 4px
        ellipse.size = (150, -2)
        r(wid)
        assert ellipse.antialiasing_line_points == []

        ellipse.size = (150, 2)
        r(wid)
        assert ellipse.antialiasing_line_points == []

        # re-enable antialiasing line rendering.
        ellipse.size = (150, 150)
        r(wid)
        assert (
            ellipse.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # disabling antialiasing.
        ellipse.texture = self._get_texture()
        r(wid)
        assert ellipse.antialiasing_line_points == []

        # re-enable antialiasing.
        ellipse.source = ""
        r(wid)
        assert (
            ellipse.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            ellipse = SmoothEllipse(
                pos=(100, 100), size=(150, 150), angle_start=90, angle_end=-120
            )
        r(wid)
        ellipse_center = [
            ellipse.pos[0] + ellipse.size[0] / 2,
            ellipse.pos[1] + ellipse.size[1] / 2,
        ]
        filtered_points = self._filtered_points(
            ellipse.points + ellipse_center
        )
        assert (
            ellipse.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # test if antialiasing is disabled if the graphic is initialized with
        # at least one of its dimensions less than 4px.
        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            ellipse = SmoothEllipse(pos=(100, 100), size=(150, -3))
        r(wid)
        assert ellipse.antialiasing_line_points == []

        wid = Widget()
        with wid.canvas:
            Color(1, 1, 1, 0.5)
            ellipse = SmoothEllipse(pos=(100, 100), size=(3.99, 3.99))
        r(wid)
        assert ellipse.antialiasing_line_points == []

    def test_smoothtriangle(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Color, SmoothTriangle

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 0, 0, 0.5)
            triangle = SmoothTriangle(
                points=[
                    100, 100,
                    200, 100,
                    150, 200,
                    500, 500, 400, 400  # Irrelevant, just for testing purposes
                ]
            )
        r(wid)
        filtered_points = self._filtered_points(triangle.points[:6])
        assert (
            triangle.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        wid = Widget()
        with wid.canvas:
            Color(0, 0, 1, 0.5)
            triangle = SmoothTriangle(
                points=[
                    125, 200,
                    200, 100,
                    100, 100,
                    500, 500, 400, 400  # Irrelevant, just for testing purposes
                ]
            )
        r(wid)
        filtered_points = self._filtered_points(triangle.points[:6])
        assert (
            triangle.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # Test disabling antialiasing if the points are very close.
        wid = Widget()
        with wid.canvas:
            Color(0, 1, 0, 0.5)
            triangle = SmoothTriangle(
                points=[100, 100, 100.5, 100, 100, 100.5]
            )
        r(wid)
        assert triangle.antialiasing_line_points == []

        # re-enable antialiasing line rendering.
        triangle.points = [125, 200, 200, 100, 100, 100]
        r(wid)
        assert (
            triangle.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # disabling antialiasing.
        triangle.texture = self._get_texture()
        r(wid)
        assert triangle.antialiasing_line_points == []

        # re-enable antialiasing.
        triangle.source = ""
        r(wid)
        assert (
            triangle.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

    def test_smoothquad(self):
        from kivy.uix.widget import Widget
        from kivy.graphics import Color, SmoothQuad

        r = self.render

        wid = Widget()
        with wid.canvas:
            Color(1, 0, 0, 0.5)
            quad = SmoothQuad(points=[100, 100, 100, 200, 200, 200, 200, 100])
        r(wid)
        filtered_points = self._filtered_points(quad.points)
        assert (
            quad.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        wid = Widget()
        with wid.canvas:
            Color(1, 0, 0, 0.5)
            quad = SmoothQuad(points=[200, 100, 200, 200, 100, 200, 100, 100])
        r(wid)
        filtered_points = self._filtered_points(quad.points)
        assert (
            quad.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # Test disabling antialiasing if the points are very close.
        wid = Widget()
        with wid.canvas:
            Color(0, 1, 0, 0.5)
            quad = SmoothQuad(
                points=[200, 100, 200, 100.8, 100, 100.8, 100, 100]
            )
        r(wid)
        assert quad.antialiasing_line_points == []

        # re-enable antialiasing line rendering.
        quad.points = [200, 100, 200, 200, 100, 200, 100, 100]
        r(wid)
        assert (
            quad.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )

        # disabling antialiasing.
        quad.texture = self._get_texture()
        r(wid)
        assert quad.antialiasing_line_points == []

        # re-enable antialiasing.
        quad.source = ""
        r(wid)
        assert (
            quad.antialiasing_line_points
            == filtered_points + filtered_points[:2]
        )


class FBOInstructionTestCase(GraphicUnitTest):

    def test_fbo_pixels(self):
        from kivy.graphics import Fbo, ClearColor, ClearBuffers, Ellipse

        fbo = Fbo(size=(512, 512))
        with fbo:
            ClearColor(0, 0, 0, 1)
            ClearBuffers()
            Ellipse(pos=(100, 100), size=(100, 100))
        fbo.draw()
        data = fbo.pixels
        fbo.texture.save('results.png')


class TransformationsTestCase(GraphicUnitTest):

    def test_identity_creation(self):
        from kivy.graphics import LoadIdentity

        mat = LoadIdentity()
        self.assertTrue(mat.stack)

    def check_transform_works(self, transform_type):
        # Normal args
        transform = transform_type(0, 1, 2)
        self.assertEqual(transform.x, 0)
        self.assertEqual(transform.y, 1)
        self.assertEqual(transform.z, 2)

        # Key word args
        transform = transform_type(x=0, y=1)
        self.assertEqual(transform.x, 0)
        self.assertEqual(transform.y, 1)

        transform = transform_type(x=0, y=1, z=2)
        self.assertEqual(transform.x, 0)
        self.assertEqual(transform.y, 1)
        self.assertEqual(transform.z, 2)

    def test_translate_creation(self):
        from kivy.graphics import Translate
        self.check_transform_works(Translate)

    def test_scale_creation(self):
        from kivy.graphics import Scale
        self.check_transform_works(Scale)


class CallbackInstructionTest(GraphicUnitTest):

    def test_from_kv(self):
        from textwrap import dedent
        from kivy.lang import Builder

        root = Builder.load_string(dedent("""\
        Widget:
            canvas:
                Callback:
                    callback: lambda __: setattr(self, 'callback_test', 'TEST')
        """))
        r = self.render
        r(root)
        self.assertTrue(root.callback_test == 'TEST')


@pytest.fixture
def widget_verify_thread(request):
    from kivy.uix.widget import Widget
    from kivy.config import Config

    original = Config.get('graphics', 'verify_gl_main_thread')
    Config.set('graphics', 'verify_gl_main_thread', request.param)

    widget = Widget()
    yield widget, request.param

    Config.set('graphics', 'verify_gl_main_thread', original)


@requires_graphics
@pytest.mark.parametrize('widget_verify_thread', ['0', '1'], indirect=True)
def test_graphics_main_thread(widget_verify_thread):
    from kivy.graphics import Color

    widget, verify_thread = widget_verify_thread
    with widget.canvas:
        color = Color()
    color.rgb = .1, .2, .3


@requires_graphics
@pytest.mark.parametrize('widget_verify_thread', ['0', '1'], indirect=True)
def test_create_graphics_second_thread(widget_verify_thread):
    from kivy.graphics import Color
    widget, verify_thread = widget_verify_thread
    exception = None

    def callback():
        nonlocal exception
        try:
            with widget.canvas:
                if verify_thread == '1':
                    with pytest.raises(TypeError):
                        Color()
                else:
                    Color()
        except BaseException as e:
            exception = e, sys.exc_info()[2]
            raise

    thread = Thread(target=callback)
    thread.start()
    thread.join()
    if exception is not None:
        raise exception[0].with_traceback(exception[1])


@requires_graphics
@pytest.mark.parametrize('widget_verify_thread', ['0', '1'], indirect=True)
def test_change_graphics_second_thread(widget_verify_thread):
    from kivy.graphics import Color
    widget, verify_thread = widget_verify_thread
    with widget.canvas:
        color = Color()

    exception = None

    def callback():
        nonlocal exception
        try:
            if verify_thread == '1':
                with pytest.raises(TypeError):
                    color.rgb = .1, .2, .3
            else:
                color.rgb = .1, .2, .3
        except BaseException as e:
            exception = e, sys.exc_info()[2]
            raise

    thread = Thread(target=callback)
    thread.start()
    thread.join()
    if exception is not None:
        raise exception[0].with_traceback(exception[1])



from kivy.app import App
from kivy.lang import Builder

kv = '''
#:import window kivy.core.window.Window

BoxLayout:
    orientation: 'vertical'
    Label:
        text: f'Window opacity: {window.opacity}'
        font_size: '25sp'
    Slider:
        size_hint_y: 4
        min: 0.0
        max: 1.0
        value: window.opacity
        on_value: window.opacity = args[1]
'''


class WindowOpacityApp(App):

    def build(self):
        return Builder.load_string(kv)


if __name__ == '__main__':
    WindowOpacityApp().run()



from itertools import product

from kivy.tests import GraphicUnitTest
from kivy.logger import LoggerHistory


class WindowBaseTest(GraphicUnitTest):

    def test_to_normalized_pos(self):
        win = self.Window
        old_system_size = win.system_size[:]
        win.system_size = w, h = type(old_system_size)((320, 240))
        try:
            for x, y in product([0, 319, 50, 51], [0, 239, 50, 51]):
                expected_sx = x / (w - 1.0)
                expected_sy = y / (h - 1.0)
                result_sx, result_sy = win.to_normalized_pos(x, y)
                assert result_sx == expected_sx
                assert result_sy == expected_sy
        finally:
            win.system_size = old_system_size


class WindowOpacityTest(GraphicUnitTest):

    def setUp(self):
        super().setUp()
        self._prev_window_opacity = self.Window.opacity
        self._prev_history = LoggerHistory.history[:]

    def tearDown(self):
        self.Window.opacity = self._prev_window_opacity
        LoggerHistory.history[:] = self._prev_history
        super().tearDown()

    def get_new_opacity_value(self):
        opacity = self.Window.opacity
        opacity = opacity - 0.1 if opacity >= 0.9 else opacity + 0.1
        return round(opacity, 2)

    def check_opacity_support(self):
        LoggerHistory.clear_history()
        self.Window.opacity = self.get_new_opacity_value()
        return not LoggerHistory.history

    def test_window_opacity_property(self):
        if self.check_opacity_support():
            opacity = self.get_new_opacity_value()
            self.Window.opacity = opacity
            self.assertEqual(self.Window.opacity, opacity)

    def test_window_opacity_clamping_positive(self):
        if self.check_opacity_support():
            self.Window.opacity = 1.5
            self.assertEqual(self.Window.opacity, 1.0)

    def test_window_opacity_clamping_negative(self):
        if self.check_opacity_support():
            self.Window.opacity = -1.5
            self.assertEqual(self.Window.opacity, 0.0)



"""
UrlRequest tests
================
"""
import threading
from base64 import b64encode
from datetime import datetime
from time import sleep

import pytest
import responses
from kivy.network.urlrequest import UrlRequestRequests as UrlRequest
from requests.auth import HTTPBasicAuth
from responses import matchers


class UrlRequestQueue:
    def __init__(self, queue):
        self.queue = queue

    def _on_success(self, req, *args):
        self.queue.append((threading.get_ident(), "success", args))

    def _on_redirect(self, req, *args):
        self.queue.append((threading.get_ident(), "redirect", args))

    def _on_error(self, req, *args):
        self.queue.append((threading.get_ident(), "error", args))

    def _on_failure(self, req, *args):
        self.queue.append((threading.get_ident(), "failure", args))

    def _on_progress(self, req, *args):
        self.queue.append((threading.get_ident(), "progress", args))

    def _on_finish(self, req, *args):
        self.queue.append((threading.get_ident(), "finish", args))


class TestCallbacks:
    url = "https://example.com"

    def _ensure_called_from_thread(self, queue):
        tid = threading.get_ident()

        for item in queue:
            assert item[0] == tid

    def _check_queue_values(self, queue_element, status):
        assert queue_element[1] == status

    def wait_request_is_finished(self, kivy_clock, request, timeout=10):
        start_time = datetime.now()
        timed_out = False

        while not request.is_finished and not timed_out:
            kivy_clock.tick()
            sleep(0.1)
            timed_out = (datetime.now() - start_time).total_seconds() > timeout

        assert request.is_finished

    @responses.activate
    def test_on_success(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=200,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_success=_queue._on_success,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "success")

    @responses.activate
    def test_on_success_with_finish(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=200,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_success=_queue._on_success,
            on_finish=_queue._on_finish,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 2
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "success")
        self._check_queue_values(processed_queue[1], "finish")

    @responses.activate
    def test_on_redirect(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=301,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_redirect=_queue._on_redirect,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "redirect")

    @responses.activate
    def test_on_redirect_with_finish(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=301,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_redirect=_queue._on_redirect,
            on_finish=_queue._on_finish,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 2
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "redirect")
        self._check_queue_values(processed_queue[1], "finish")

    @responses.activate
    def test_on_error(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body=Exception("..."),
            status=400,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_error=_queue._on_error,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "error")

    @responses.activate
    def test_on_error_with_finis(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body=Exception("..."),
            status=400,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_error=_queue._on_error,
            on_finish=_queue._on_finish,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 2
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "error")
        self._check_queue_values(processed_queue[1], "finish")

    @responses.activate
    def test_on_failure(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=400,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_failure=_queue._on_failure,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "failure")

    @responses.activate
    def test_on_failure_with_finish(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=400,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_failure=_queue._on_failure,
            on_finish=_queue._on_finish,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 2
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "failure")
        self._check_queue_values(processed_queue[1], "finish")

    @responses.activate
    def test_on_progress(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="x" * 100,
            status=200,
            content_type="text/plain",
            auto_calculate_content_length=True
        )

        req = UrlRequest(
            self.url,
            on_progress=_queue._on_progress,
            chunk_size=70,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 4
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "progress")
        self._check_queue_values(processed_queue[1], "progress")
        self._check_queue_values(processed_queue[2], "progress")
        self._check_queue_values(processed_queue[3], "progress")

    @responses.activate
    def test_on_progress_with_finish(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="x" * 100,
            status=200,
            content_type="text/plain",
            auto_calculate_content_length=True
        )

        req = UrlRequest(
            self.url,
            on_progress=_queue._on_progress,
            on_finish=_queue._on_finish,
            chunk_size=70,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 5
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "progress")
        self._check_queue_values(processed_queue[1], "progress")
        self._check_queue_values(processed_queue[2], "progress")
        self._check_queue_values(processed_queue[3], "progress")
        self._check_queue_values(processed_queue[4], "finish")

    @responses.activate
    def test_on_finish(self, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            self.url,
            body="{}",
            status=400,
            content_type="application/json",
        )

        req = UrlRequest(
            self.url,
            on_finish=_queue._on_finish,
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "finish")

    @responses.activate
    def test_auth_header(self, kivy_clock):
        _queue = UrlRequestQueue([])
        head = {
            "Authorization": "Basic {}".format(
                b64encode(b"exampleuser:examplepassword").decode("utf-8")
            )
        }
        responses.get(
            self.url,
            body="{}",
            status=400,
            content_type="application/json",
            match=[matchers.header_matcher(head)],
        )

        req = UrlRequest(
            self.url,
            req_headers=head,
            on_finish=_queue._on_finish,
            debug=True,
            auth=HTTPBasicAuth("exampleuser", "examplepassword")
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "finish")

    @pytest.mark.parametrize("scheme", ("http", "https"))
    @responses.activate
    def test_ca_file(self, scheme, kivy_clock):
        _queue = UrlRequestQueue([])
        responses.get(
            f"{scheme}://example.com",
            body="{}",
            status=400,
            content_type="application/json",
        )

        req = UrlRequest(
            f"{scheme}://example.com",
            on_finish=_queue._on_finish,
            ca_file="/etc/ssl/certs/ca-certificates.crt",
            debug=True,
        )
        self.wait_request_is_finished(kivy_clock, req)

        processed_queue = _queue.queue
        assert len(processed_queue) == 1
        self._ensure_called_from_thread(processed_queue)
        self._check_queue_values(processed_queue[0], "finish")



"""
Kivy Logging
============

By default, Kivy provides a logging system based on the standard Python
`logging <https://docs.python.org/3/library/logging.html>`_ module with
several additional features designed to be more convenient. These features
include:

 * simplied usage (single instance, simple configuration, works by default)
 * color-coded output on supported terminals
 * output to ``stderr`` by default
 * message categorization via colon separation
 * access to log history even if logging is disabled
 * built-in handling of various cross-platform considerations
 * any stray output written to ``sys.stderr`` is captured, and stored in the log
   file as a warning.

These features are configurable via the Config file or environment variables -
including falling back to only using the standard Python system.

Logger object
=============

The Kivy ``Logger`` class provides a singleton logging.logger instance.

As well as the standard logging levels (``debug``, ``info``,
``warning``, ``error`` and ``critical``), an additional ``trace`` level is
available.

Example Usage
-------------

Use the ``Logger`` as you would a standard Python logger. ::

    from kivy.logger import Logger

    Logger.info('title: This is a info message.')
    Logger.debug('title: This is a debug message.')

    try:
        raise Exception('bleh')
    except Exception:
        Logger.exception('Something happened!')

The message passed to the logger is split into two parts separated by a colon
(:). The first part is used as a title and the second part is used as the
message. This way, you can "categorize" your messages easily. ::

    Logger.info('Application: This is a test')

    # will appear as

    [INFO   ] [Application ] This is a test

You can change the logging level at any time using the ``setLevel`` method. ::

    from kivy.logger import Logger, LOG_LEVELS

    Logger.setLevel(LOG_LEVELS["debug"])

.. versionchanged:: 2.2.0

Interaction with other logging
------------------------------

The Kivy logging system will, by default, present all log messages sent from
any logger - e.g. from third-party libraries.

Additional handlers may be added.

.. warning:: Handlers that output to ``sys.stderr`` may cause loops, as stderr
   output is reported as a warning log message.

Logger Configuration
====================

Kivy Log Mode
-------------

At the highest level, Kivy's logging system is controlled by an environment
variable ``KIVY_LOG_MODE``. It may be given any of three values:
``KIVY``, ``PYTHON``, ``MIXED``

.. versionadded: 2.2.0

KIVY Mode (default)
^^^^^^^^^^^^^^^^^^^

In ``KIVY`` mode, all Kivy handlers are attached to the root logger, so all log
messages in the system are output to the Kivy log files and to the console. Any
stray output to ``sys.stderr`` is logged as a warning.

If you are writing an entire Kivy app from scratch, this is the most convenient
mode.

PYTHON Mode
^^^^^^^^^^^

In ``PYTHON`` mode, no handlers are added, and ``sys.stderr`` output is not
captured. It is left to the client to add appropriate handlers. (If none are
added, the ``logging`` module will output them to ``stderr``.)

Messages logged with ``Logger`` will be propagated to the root logger, from a
logger named ``kivy``.

If the Kivy app is part of a much larger project which has its own logging
regimen, this is the mode that gives most control.

The ``kivy.logger`` file contains a number of ``logging.handler``,
``logging.formatter``, and other helper classes to allow
users to adopt the features of Kivy logging that they like, including the
stderr redirection.

MIXED Mode
^^^^^^^^^^

In ``MIXED`` mode, handlers are added to the Kivy's ``Logger`` object directly,
and propagation is turned off. ``sys.stderr`` is not redirected.

Messages logged with Kivy's ``Logger`` will appear in the Kivy log file and
output to the Console.

However, messages logged with other Python loggers will not be handled by Kivy
handlers. The client will need to add their own.

If you like the features of Kivy ``Logger``, but are writing a Kivy app that
relies on third-party libraries that don't use colon-separation of categorise
or depend on the display of the logger name, this mode provides a compromise.

Again, the ``kivy.logger`` file contains re-usable logging features that can be
used to get the best of both systems.

Config Files
------------

In ``KIVY`` and ``MIXED`` modes, the logger handlers can be controlled via the
Kivy configuration file::

    [kivy]
    log_level = info
    log_enable = 1
    log_dir = logs
    log_name = kivy_%y-%m-%d_%_.txt
    log_maxfiles = 100

More information about the allowed values are described in the
:mod:`kivy.config` module.

In addition, the environment variables ``KIVY_NO_FILELOG`` and
``KIVY_NO_CONSOLELOG`` can be used to turn off the installation of the
corresponding handlers.


Logger History
--------------

Even if the logger is not enabled, you still have access to the last 100
LogRecords::

    from kivy.logger import LoggerHistory

    print(LoggerHistory.history)
"""

import logging
import os
import sys
from functools import partial
import pathlib

import kivy
from kivy.utils import platform


__all__ = (
    "add_kivy_handlers",
    "ColonSplittingLogRecord",
    "ColoredLogRecord",
    "COLORS",
    "ConsoleHandler",
    "file_log_handler",
    "FileHandler",
    "is_color_terminal",
    "KivyFormatter",
    "LOG_LEVELS",
    "Logger",
    "LoggerHistory",
    "ProcessingStream",
    "UncoloredLogRecord",
)


Logger = None


logging.addLevelName(9, 'TRACE')
logging.TRACE = 9
LOG_LEVELS = {
    'trace': logging.TRACE,
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL}


class FileHandler(logging.Handler):
    history = []
    filename = 'log.txt'
    fd = None
    log_dir = ''
    encoding = 'utf-8'

    def purge_logs(self):
        """Purge logs which exceed the maximum amount of log files,
        starting with the oldest creation timestamp (or edit-timestamp on Linux)
        """

        if not self.log_dir:
            return

        from kivy.config import Config
        maxfiles = Config.getint("kivy", "log_maxfiles")

        # Get path to log directory
        log_dir = pathlib.Path(self.log_dir)

        if maxfiles < 0:  # No log file limit set
            return

        Logger.info("Logger: Purge log fired. Processing...")

        # Get all files from log directory and corresponding creation timestamps
        files = [(item, item.stat().st_ctime)
                 for item in log_dir.iterdir() if item.is_file()]
        # Sort files by ascending timestamp
        files.sort(key=lambda x: x[1])

        for file, _ in files[:(-maxfiles or len(files))]:
            # More log files than allowed maximum,
            # delete files, starting with oldest creation timestamp
            # (or edit-timestamp on Linux)
            try:
                file.unlink()
            except (PermissionError, FileNotFoundError) as e:
                Logger.info(f"Logger: Skipped file {file}, {repr(e)}")

        Logger.info("Logger: Purge finished!")

    def _configure(self, *largs, **kwargs):
        from time import strftime
        from kivy.config import Config
        log_dir = Config.get('kivy', 'log_dir')
        log_name = Config.get('kivy', 'log_name')

        _dir = kivy.kivy_home_dir
        if log_dir and os.path.isabs(log_dir):
            _dir = log_dir
        else:
            _dir = os.path.join(_dir, log_dir)
        if not os.path.exists(_dir):
            os.makedirs(_dir)
        self.log_dir = _dir

        pattern = log_name.replace('%_', '@@NUMBER@@')
        pattern = os.path.join(_dir, strftime(pattern))
        n = 0
        while True:
            filename = pattern.replace('@@NUMBER@@', str(n))
            if not os.path.exists(filename):
                break
            n += 1
            if n > 10000:  # prevent maybe flooding ?
                raise Exception('Too many logfile, remove them')

        if FileHandler.filename == filename and FileHandler.fd is not None:
            return

        FileHandler.filename = filename
        if FileHandler.fd not in (None, False):
            FileHandler.fd.close()
        FileHandler.fd = open(filename, 'w', encoding=FileHandler.encoding)
        Logger.info('Logger: Record log in %s' % filename)

    def _write_message(self, record):
        if FileHandler.fd in (None, False):
            return

        msg = self.format(record)
        stream = FileHandler.fd
        fs = "%s\n"
        stream.write('[%-7s] ' % record.levelname)
        stream.write(fs % msg)
        stream.flush()

    def emit(self, message):
        # during the startup, store the message in the history
        if Logger.logfile_activated is None:
            FileHandler.history += [message]
            return

        # startup done, if the logfile is not activated, avoid history.
        if Logger.logfile_activated is False:
            FileHandler.history = []
            return

        if FileHandler.fd is None:
            try:
                self._configure()
                from kivy.config import Config
                Config.add_callback(self._configure, 'kivy', 'log_dir')
                Config.add_callback(self._configure, 'kivy', 'log_name')
            except Exception:
                # deactivate filehandler...
                if FileHandler.fd not in (None, False):
                    FileHandler.fd.close()
                FileHandler.fd = False
                Logger.exception('Error while activating FileHandler logger')
                return
            while FileHandler.history:
                _message = FileHandler.history.pop()
                self._write_message(_message)

        self._write_message(message)


class LoggerHistory(logging.Handler):

    history = []

    def emit(self, message):
        LoggerHistory.history = [message] + LoggerHistory.history[:100]

    @classmethod
    def clear_history(cls):
        del cls.history[:]

    def flush(self):
        super(LoggerHistory, self).flush()
        self.clear_history()


class ConsoleHandler(logging.StreamHandler):
    """
        Emits records to a stream (by default, stderr).

        However, if the msg starts with "stderr:" it is not formatted, but
        written straight to the stream.

        .. versionadded:: 2.2.0
    """

    def filter(self, record):
        try:
            msg = record.msg
            k = msg.split(':', 1)
            if k[0] == 'stderr' and len(k) == 2:
                # This message was scraped from stderr.
                # Emit it without formatting.
                self.stream.write(k[1] + '\n')
                # Don't pass it to the formatted emitter.
                return False
        except Exception:
            pass
        return True


class ProcessingStream(object):
    """
        Stream-like object that takes each completed line written to it,
        adds a given prefix, and applies the given function to it.

        .. versionadded:: 2.2.0
    """

    def __init__(self, channel, func):
        self.buffer = ""
        self.func = func
        self.channel = channel
        self.errors = ""

    def write(self, s):
        s = self.buffer + s
        self.flush()
        f = self.func
        channel = self.channel
        lines = s.split('\n')
        for line in lines[:-1]:
            f('%s: %s' % (channel, line))
        self.buffer = lines[-1]

    def flush(self):
        return

    def isatty(self):
        return False


def logger_config_update(section, key, value):
    if KIVY_LOG_MODE != "PYTHON":
        if LOG_LEVELS.get(value) is None:
            raise AttributeError('Loglevel {0!r} doesn\'t exists'.format(value))
        Logger.setLevel(level=LOG_LEVELS.get(value))


class ColonSplittingLogRecord(logging.LogRecord):
    """Clones an existing logRecord, but reformats the message field
    if it contains a colon.

    .. versionadded:: 2.2.0
    """

    def __init__(self, logrecord):
        try:
            parts = logrecord.msg.split(":", 1)
            if len(parts) == 2:
                new_msg = "[%-12s]%s" % (parts[0], parts[1])
            else:
                new_msg = parts[0]
        except Exception:
            new_msg = logrecord.msg

        super().__init__(
            name=logrecord.name,
            level=logrecord.levelno,
            pathname=logrecord.pathname,
            lineno=logrecord.lineno,
            msg=new_msg,
            args=logrecord.args,
            exc_info=logrecord.exc_info,
            func=logrecord.funcName,
            sinfo=logrecord.stack_info,
        )


class ColoredLogRecord(logging.LogRecord):
    """Clones an existing logRecord, but reformats the levelname to add
    color, and the message to add bolding (where indicated by $BOLD
    and $RESET in the message).

    .. versionadded:: 2.2.0"""

    BLACK = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    MAGENTA = 5
    CYAN = 6
    WHITE = 7
    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"
    BOLD_SEQ = "\033[1m"

    LEVEL_COLORS = {
        "TRACE": MAGENTA,
        "WARNING": YELLOW,
        "INFO": GREEN,
        "DEBUG": CYAN,
        "CRITICAL": RED,
        "ERROR": RED,
    }

    @classmethod
    def _format_message(cls, message):
        return str(message).replace(
            "$RESET", cls.RESET_SEQ).replace("$BOLD", cls.BOLD_SEQ)

    @classmethod
    def _format_levelname(cls, levelname):
        if levelname in cls.LEVEL_COLORS:
            return (
                cls.COLOR_SEQ % (30 + cls.LEVEL_COLORS[levelname])
                + levelname
                + cls.RESET_SEQ
            )
        return levelname

    def __init__(self, logrecord):
        super().__init__(
            name=logrecord.name,
            level=logrecord.levelno,
            pathname=logrecord.pathname,
            lineno=logrecord.lineno,
            msg=logrecord.msg,
            args=logrecord.args,
            exc_info=logrecord.exc_info,
            func=logrecord.funcName,
            sinfo=logrecord.stack_info,
        )
        self.levelname = self._format_levelname(self.levelname)
        self.msg = self._format_message(self.msg)


# Included for backward compatibility only.
# Could be used to override colors.
COLORS = ColoredLogRecord.LEVEL_COLORS


class UncoloredLogRecord(logging.LogRecord):
    """Clones an existing logRecord, but reformats the message
    to remove $BOLD/$RESET markup.

    .. versionadded:: 2.2.0"""

    @classmethod
    def _format_message(cls, message):
        return str(message).replace("$RESET", "").replace("$BOLD", "")

    def __init__(self, logrecord):
        super().__init__(
            name=logrecord.name,
            level=logrecord.levelno,
            pathname=logrecord.pathname,
            lineno=logrecord.lineno,
            msg=logrecord.msg,
            args=logrecord.args,
            exc_info=logrecord.exc_info,
            func=logrecord.funcName,
            sinfo=logrecord.stack_info,
        )
        self.msg = self._format_message(self.msg)


class KivyFormatter(logging.Formatter):
    """Split out first field in message marked with a colon,
    and either apply terminal color codes to the record, or strip
    out color markup if colored logging is not available.

    .. versionadded:: 2.2.0"""

    def __init__(self, *args, use_color=True, **kwargs):
        super().__init__(*args, **kwargs)
        self._coloring_cls = (
            ColoredLogRecord if use_color else UncoloredLogRecord)

    def format(self, record):
        return super().format(
            self._coloring_cls(ColonSplittingLogRecord(record)))


def is_color_terminal():
    """ Detect whether the environment supports color codes in output.

    .. versionadded:: 2.2.0
    """

    return (
            (
                    os.environ.get("WT_SESSION") or
                    os.environ.get("COLORTERM") == 'truecolor' or
                    os.environ.get('PYCHARM_HOSTED') == '1' or
                    os.environ.get('TERM') in (
                        'rxvt',
                        'rxvt-256color',
                        'rxvt-unicode',
                        'rxvt-unicode-256color',
                        'xterm',
                        'xterm-256color',
                    )
            )
            and platform not in ('android', 'ios')
    )


#: Kivy default logger instance
# .. versionchanged:: 2.2.0
Logger = logging.getLogger('kivy')
Logger.logfile_activated = None
Logger.trace = partial(Logger.log, logging.TRACE)

file_log_handler = (
    FileHandler()
    if 'KIVY_NO_FILELOG' not in os.environ
    else None
)


# Issue #7891 describes an undocumented feature that was since removed
# Detect if a client was depending on it.
# .. versionchanged:: 2.2.0
assert not hasattr(sys, '_kivy_logging_handler'), \
    "Not supported. Try logging.root.addHandler()"


def add_kivy_handlers(logger):
    """ Add Kivy-specific handlers to a logger.

    .. versionadded:: 2.2.0
    """
    # add default kivy logger
    logger.addHandler(LoggerHistory())
    if file_log_handler:
        logger.addHandler(file_log_handler)

    # Use the custom handler instead of streaming one.
    # Don't output to stderr if it is set to None
    # stderr is set to None by pythonw and pyinstaller 5.7+
    if sys.stderr and 'KIVY_NO_CONSOLELOG' not in os.environ:
        use_color = is_color_terminal()
        if not use_color:
            # No additional control characters will be inserted inside the
            # levelname field, 7 chars will fit "WARNING"
            fmt = "[%(levelname)-7s] %(message)s"
        else:
            # levelname field width need to take into account the length of
            # the color control codes (7+4 chars for bold+color, and reset)
            fmt = "[%(levelname)-18s] %(message)s"
        formatter = KivyFormatter(fmt, use_color=use_color)
        console = ConsoleHandler()
        console.setFormatter(formatter)
        logger.addHandler(console)


KIVY_LOG_MODE = os.environ.get("KIVY_LOG_MODE", "KIVY")
assert KIVY_LOG_MODE in ("KIVY", "PYTHON", "MIXED"), "Unknown log mode"

if KIVY_LOG_MODE == "KIVY":
    # Add the Kivy handlers to the root logger, so they will be used
    # for all propagated log messages.
    add_kivy_handlers(logging.root)

    # Root logger defaults to warning. Let Logger be the limiting factor.
    logging.root.setLevel(logging.NOTSET)

    # install stderr handlers

    # Caution: If any logging handlers output to sys.stderr they should be
    # configured BEFORE this reconfiguration is done to avoid loops.
    sys.stderr = ProcessingStream("stderr", Logger.warning)
    # Sends all messages written to stderr to the Logger, after prefixing it
    # with "stderr:"
elif KIVY_LOG_MODE == "MIXED":
    # Add the Kivy handlers to the Kivy logger, so they will be used
    # for all messages sent through Logger, only.
    add_kivy_handlers(Logger)

    # Don't spread Kivy-related log messages to the root logger.
    Logger.propagate = False

    # Don't set stderr redirection: it is too likely to cause loops with other
    # handlers. Client can manually add it, if desired.
else:  # KIVY_LOG_MODE == "PYTHON"
    # Don't add handlers or redirect stderr. Client can manually add if desired.
    pass



'''
Bubble
======

.. versionadded:: 1.1.0

.. image:: images/bubble.jpg
    :align: right

The :class:`Bubble` widget is a form of menu or a small popup with an arrow
arranged on one side of it's content.

The :class:`Bubble` contains an arrow attached to the content
(e.g., :class:`BubbleContent`) pointing in the direction you choose. It can
be placed either at a predefined location or flexibly by specifying a relative
position on the border of the widget.

The :class:`BubbleContent` is a styled BoxLayout and is thought to be added to
the :class:`Bubble` as a child widget. The :class:`Bubble` will then arrange
an arrow around the content as desired. Instead of the class:`BubbleContent`,
you can theoretically use any other :class:`Widget` as well as long as it
supports the 'bind' and 'unbind' function of the :class:`EventDispatcher` and
is compatible with Kivy to be placed inside a :class:`BoxLayout`.

The :class:`BubbleButton`is a styled Button. It suits to the style of
:class:`Bubble` and :class:`BubbleContent`. Feel free to place other Widgets
inside the 'content' of the :class:`Bubble`.


.. versionchanged:: 2.2.0
The properties :attr:`background_image`, :attr:`background_color`,
:attr:`border` and :attr:`border_auto_scale` were removed from :class:`Bubble`.
These properties had only been used by the content widget that now uses it's
own properties instead. The color of the arrow is now changed with
:attr:`arrow_color` instead of :attr:`background_color`.
These changes makes the :class:`Bubble` transparent to use with other layouts
as content without any side-effects due to property inheritance.

The property :attr:`flex_arrow_pos` has been added to allow further
customization of the arrow positioning.

The properties :attr:`arrow_margin`, :attr:`arrow_margin_x`,
:attr:`arrow_margin_y`, :attr:`content_size`, :attr:`content_width` and
:attr:`content_height` have been added to ease proper sizing of a
:class:`Bubble` e.g., based on it's content size.

BubbleContent
=============

The :class:`BubbleContent` is a styled BoxLayout that can be used to
add e.g., :class:`BubbleButtons` as menu items.

.. versionchanged:: 2.2.0
The properties :attr:`background_image`, :attr:`background_color`,
:attr:`border` and :attr:`border_auto_scale` were added to the
:class:`BubbleContent`. The :class:`BubbleContent` does no longer rely on these
properties being present in the parent class.

BubbleButton
============

The :class:`BubbleButton` is a styled :class:`Button` that can be used to be
added to the :class:`BubbleContent`.

Simple example
--------------

.. include:: ../../examples/widgets/bubble_test.py
    :literal:

Customize the Bubble
--------------------

You can choose the direction in which the arrow points::

    Bubble(arrow_pos='top_mid')
    or
    Bubble(size=(200, 40), flex_arrow_pos=(175, 40))

    Similarly, the corresponding properties in the '.kv' language can be used
    as well.

You can change the appearance of the bubble::

    Bubble(
        arrow_image='/path/to/arrow/image',
        arrow_color=(1, 0, 0, .5)),
    )
    BubbleContent(
        background_image='/path/to/background/image',
        background_color=(1, 0, 0, .5),  # 50% translucent red
        border=(0,0,0,0),
    )

    Similarly, the corresponding properties in the '.kv' language can be used
    as well.

-----------------------------
'''

__all__ = ('Bubble', 'BubbleButton', 'BubbleContent')

from kivy.uix.image import Image
from kivy.uix.scatter import Scatter
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.button import Button
from kivy.properties import ObjectProperty
from kivy.properties import StringProperty
from kivy.properties import OptionProperty
from kivy.properties import ListProperty
from kivy.properties import BooleanProperty
from kivy.properties import ColorProperty
from kivy.properties import NumericProperty
from kivy.properties import ReferenceListProperty
from kivy.base import EventLoop
from kivy.metrics import dp


class BubbleException(Exception):
    pass


class BubbleButton(Button):
    '''A button intended for use in a BubbleContent widget.
    You can use a "normal" button class, but it will not look good unless the
    background is changed.

    Rather use this BubbleButton widget that is already defined and provides a
    suitable background for you.
    '''
    pass


class BubbleContent(BoxLayout):
    '''A styled BoxLayout that can be used as the content widget of a Bubble.

    .. versionchanged:: 2.2.0
    The graphical appearance of :class:`BubbleContent` is now based on it's
    own properties :attr:`background_image`, :attr:`background_color`,
    :attr:`border` and :attr:`border_auto_scale`. The parent widget properties
    are no longer considered. This makes the BubbleContent a standalone themed
    BoxLayout.
    '''

    background_color = ColorProperty([1, 1, 1, 1])
    '''Background color, in the format (r, g, b, a). To use it you have to set
    :attr:`background_image` first.

    .. versionadded:: 2.2.0

    :attr:`background_color` is a :class:`~kivy.properties.ColorProperty` and
    defaults to [1, 1, 1, 1].
    '''

    background_image = StringProperty('atlas://data/images/defaulttheme/bubble')
    '''Background image of the bubble.

    .. versionadded:: 2.2.0

    :attr:`background_image` is a :class:`~kivy.properties.StringProperty` and
    defaults to 'atlas://data/images/defaulttheme/bubble'.
    '''

    border = ListProperty([16, 16, 16, 16])
    '''Border used for :class:`~kivy.graphics.vertex_instructions.BorderImage`
    graphics instruction. Used with the :attr:`background_image`.
    It should be used when using custom backgrounds.

    It must be a list of 4 values: (bottom, right, top, left). Read the
    BorderImage instructions for more information about how to use it.

    .. versionadded:: 2.2.0

    :attr:`border` is a :class:`~kivy.properties.ListProperty` and defaults to
    (16, 16, 16, 16)
    '''

    border_auto_scale = OptionProperty(
        'both_lower',
        options=[
            'off', 'both', 'x_only', 'y_only', 'y_full_x_lower',
            'x_full_y_lower', 'both_lower'
        ]
    )
    '''Specifies the :attr:`kivy.graphics.BorderImage.auto_scale`
    value on the background BorderImage.

    .. versionadded:: 2.2.0

    :attr:`border_auto_scale` is a
    :class:`~kivy.properties.OptionProperty` and defaults to
    'both_lower'.
    '''


class Bubble(BoxLayout):
    '''Bubble class. See module documentation for more information.
    '''

    content = ObjectProperty(allownone=True)
    '''This is the object where the main content of the bubble is held.

    The content of the Bubble set by 'add_widget' and removed with
    'remove_widget' similarly to the :class:`ActionView` which is placed into
    a class:`ActionBar`

    :attr:`content` is a :class:`~kivy.properties.ObjectProperty` and defaults
    to None.
    '''

    arrow_image = StringProperty(
        'atlas://data/images/defaulttheme/bubble_arrow'
    )
    ''' Image of the arrow pointing to the bubble.

    :attr:`arrow_image` is a :class:`~kivy.properties.StringProperty` and
    defaults to 'atlas://data/images/defaulttheme/bubble_arrow'.
    '''

    arrow_color = ColorProperty([1, 1, 1, 1])
    '''Arrow color, in the format (r, g, b, a). To use it you have to set
    :attr:`arrow_image` first.

    .. versionadded:: 2.2.0

    :attr:`arrow_color` is a :class:`~kivy.properties.ColorProperty` and
    defaults to [1, 1, 1, 1].
    '''

    show_arrow = BooleanProperty(True)
    ''' Indicates whether to show arrow.

    .. versionadded:: 1.8.0

    :attr:`show_arrow` is a :class:`~kivy.properties.BooleanProperty` and
    defaults to `True`.
    '''

    arrow_pos = OptionProperty(
        'bottom_mid',
        options=(
            'left_top', 'left_mid', 'left_bottom',
            'top_left', 'top_mid', 'top_right',
            'right_top', 'right_mid', 'right_bottom',
            'bottom_left', 'bottom_mid', 'bottom_right',
        )
    )
    '''Specifies the position of the arrow as predefined relative position to
    the bubble.
    Can be one of: left_top, left_mid, left_bottom top_left, top_mid, top_right
    right_top, right_mid, right_bottom bottom_left, bottom_mid, bottom_right.

    :attr:`arrow_pos` is a :class:`~kivy.properties.OptionProperty` and
    defaults to 'bottom_mid'.
    '''

    flex_arrow_pos = ListProperty(None)
    '''Specifies the position of the arrow as flex coordinate around the
    border of the :class:`Bubble` Widget.
    If this property is set to a proper position (relative pixel coordinates
    within the :class:`Bubble` widget, it overwrites the setting
    :attr:`arrow_pos`.

    .. versionadded:: 2.2.0

    :attr:`flex_arrow_pos` is a :class:`~kivy.properties.ListProperty` and
    defaults to None.
    '''

    limit_to = ObjectProperty(None, allownone=True)
    '''Specifies the widget to which the bubbles position is restricted.

    .. versionadded:: 1.6.0

    :attr:`limit_to` is a :class:`~kivy.properties.ObjectProperty` and defaults
    to 'None'.
    '''

    arrow_margin_x = NumericProperty(0)
    '''Automatically computed margin in x direction that the arrow widget
    occupies in pixel.

    In combination with the :attr:`content_width`, this property can be used
    to determine the correct width of the Bubble to exactly enclose the
    arrow + content by adding :attr:`content_width` and :attr:`arrow_margin_x`

    .. versionadded:: 2.2.0

    :attr:`arrow_margin_x` is a :class:`~kivy.properties.NumericProperty` and
    represents the added margin in x direction due to the arrow widget.
    It defaults to 0 and is read only.
    '''

    arrow_margin_y = NumericProperty(0)
    '''Automatically computed margin in y direction that the arrow widget
    occupies in pixel.

    In combination with the :attr:`content_height`, this property can be used
    to determine the correct height of the Bubble to exactly enclose the
    arrow + content by adding :attr:`content_height` and :attr:`arrow_margin_y`

    .. versionadded:: 2.2.0

    :attr:`arrow_margin_y` is a :class:`~kivy.properties.NumericProperty` and
    represents the added margin in y direction due to the arrow widget.
    It defaults to 0 and is read only.
    '''

    arrow_margin = ReferenceListProperty(arrow_margin_x, arrow_margin_y)
    '''Automatically computed margin that the arrow widget occupies in
    x and y direction in pixel.

    Check the description of :attr:`arrow_margin_x` and :attr:`arrow_margin_y`.

    .. versionadded:: 2.2.0

    :attr:`arrow_margin` is a :class:`~kivy.properties.ReferenceListProperty`
    of (:attr:`arrow_margin_x`, :attr:`arrow_margin_y`) properties.
    It is read only.
    '''

    content_width = NumericProperty(0)
    '''The width of the content Widget.

    .. versionadded:: 2.2.0

    :attr:`content_width` is a :class:`~kivy.properties.NumericProperty` and
    is the same as self.content.width if content is not None, else it defaults
    to 0. It is read only.
    '''

    content_height = NumericProperty(0)
    '''The height of the content Widget.

    .. versionadded:: 2.2.0

    :attr:`content_height` is a :class:`~kivy.properties.NumericProperty` and
    is the same as self.content.height if content is not None, else it defaults
    to 0. It is read only.
    '''

    content_size = ReferenceListProperty(content_width, content_height)
    ''' The size of the content Widget.

    .. versionadded:: 2.2.0

    :attr:`content_size` is a :class:`~kivy.properties.ReferenceListProperty`
    of (:attr:`content_width`, :attr:`content_height`) properties.
    It is read only.
    '''

    # Internal map that specifies the different parameters for fixed arrow
    # position layouts. The flex_arrow_pos uses these parameter sets
    # as a template.
    # 0: orientation of the children of Bubble ([content, arrow])
    # 1: order of widgets to add to the BoxLayout (default: [content, arrow])
    # 2: size_hint of _arrow_image_layout
    # 3: rotation of the _arrow_image
    # 4: pos_hint of the _arrow_image_layout
    ARROW_LAYOUTS = {
        "bottom_left":  (   "vertical",  1, (   1, None),   0, {   "top": 1.0,        "x": 0.05}),  # noqa: E201,E241,E501
        "bottom_mid":   (   "vertical",  1, (   1, None),   0, {   "top": 1.0, "center_x": 0.50}),  # noqa: E201,E241,E501
        "bottom_right": (   "vertical",  1, (   1, None),   0, {   "top": 1.0,    "right": 0.95}),  # noqa: E201,E241,E501
        "right_bottom": ( "horizontal",  1, (None,    1),  90, {  "left": 0.0,        "y": 0.05}),  # noqa: E201,E241,E501
        "right_mid":    ( "horizontal",  1, (None,    1),  90, {  "left": 0.0, "center_y": 0.50}),  # noqa: E201,E241,E501
        "right_top":    ( "horizontal",  1, (None,    1),  90, {  "left": 0.0,      "top": 0.95}),  # noqa: E201,E241,E501
        "top_left":     (   "vertical", -1, (   1, None), 180, {"bottom": 0.0,        "x": 0.05}),  # noqa: E201,E241,E501
        "top_mid":      (   "vertical", -1, (   1, None), 180, {"bottom": 0.0, "center_x": 0.50}),  # noqa: E201,E241,E501
        "top_right":    (   "vertical", -1, (   1, None), 180, {"bottom": 0.0,    "right": 0.95}),  # noqa: E201,E241,E501
        "left_bottom":  ( "horizontal", -1, (None,    1), -90, {"right":  1.0,        "y": 0.05}),  # noqa: E201,E241,E501
        "left_mid":     ( "horizontal", -1, (None,    1), -90, {"right":  1.0, "center_y": 0.50}),  # noqa: E201,E241,E501
        "left_top":     ( "horizontal", -1, (None,    1), -90, {"right":  1.0,      "top": 0.95}),  # noqa: E201,E241,E501
    }

    def __init__(self, **kwargs):
        self.content = None

        self._flex_arrow_layout_params = None
        self._temporarily_ignore_limits = False

        self._arrow_image = Image(
            source=self.arrow_image,
            fit_mode="scale-down",
            color=self.arrow_color
        )
        self._arrow_image.width = self._arrow_image.texture_size[0]
        self._arrow_image.height = dp(self._arrow_image.texture_size[1])
        self._arrow_image_scatter = Scatter(
            size_hint=(None, None),
            do_scale=False,
            do_rotation=False,
            do_translation=False,
        )
        self._arrow_image_scatter.add_widget(self._arrow_image)
        self._arrow_image_scatter.size = self._arrow_image.texture_size
        self._arrow_image_scatter_wrapper = BoxLayout(
            size_hint=(None, None),
        )
        self._arrow_image_scatter_wrapper.add_widget(self._arrow_image_scatter)
        self._arrow_image_layout = RelativeLayout()
        self._arrow_image_layout.add_widget(self._arrow_image_scatter_wrapper)

        self._arrow_layout = None

        super().__init__(**kwargs)
        self.reposition_inner_widgets()

    def add_widget(self, widget, *args, **kwargs):
        if self.content is None:
            self.content = widget
            self.content_size = widget.size
            self.content.bind(size=self.update_content_size)
            self.reposition_inner_widgets()
        else:
            raise BubbleException(
                "Bubble can only contain a single Widget or Layout"
            )

    def remove_widget(self, widget, *args, **kwargs):
        if widget == self.content:
            self.content.unbind(size=self.update_content_size)
            self.content = None
            self.content_size = [0, 0]
            self.reposition_inner_widgets()
            return
        super().remove_widget(widget, *args, **kwargs)

    def on_content_size(self, instance, value):
        self.adjust_position()

    def on_limit_to(self, instance, value):
        self.adjust_position()

    def on_pos(self, instance, value):
        self.adjust_position()

    def on_size(self, instance, value):
        self.reposition_inner_widgets()

    def on_arrow_image(self, instance, value):
        self._arrow_image.source = self.arrow_image
        self._arrow_image.width = self._arrow_image.texture_size[0]
        self._arrow_image.height = dp(self._arrow_image.texture_size[1])
        self._arrow_image_scatter.size = self._arrow_image.texture_size
        self.reposition_inner_widgets()

    def on_arrow_color(self, instance, value):
        self._arrow_image.color = self.arrow_color

    def on_arrow_pos(self, instance, value):
        self.reposition_inner_widgets()

    def on_flex_arrow_pos(self, instance, value):
        self._flex_arrow_layout_params = self.get_flex_arrow_layout_params()
        self.reposition_inner_widgets()

    def get_flex_arrow_layout_params(self):
        pos = self.flex_arrow_pos

        if pos is None:
            return None

        x, y = pos
        if not (0 <= x <= self.width and 0 <= y <= self.height):
            return None

        # the order of the following list defines the side that the arrow
        # will be attached to in case of ambiguity (same distances)
        base_layouts_map = [
            ("bottom_mid", y),
            ("top_mid", self.height - y),
            ("left_mid", x),
            ("right_mid", self.width - x),
        ]
        base_layout_key = min(base_layouts_map, key=lambda val: val[1])[0]
        arrow_layout = list(Bubble.ARROW_LAYOUTS[base_layout_key])

        arrow_width = self._arrow_image.width

        # This function calculates the proper value for pos_hint, i.e., the
        # arrow texture does not 'overflow' and stays entirely connected to
        # the side of the content.
        def calc_x0(x, length):
            return x * (length - arrow_width) / (length * length)

        if base_layout_key == "bottom_mid":
            arrow_layout[-1] = {"top": 1.0, "x": calc_x0(x, self.width)}
        elif base_layout_key == "top_mid":
            arrow_layout[-1] = {"bottom": 0.0, "x": calc_x0(x, self.width)}
        elif base_layout_key == "left_mid":
            arrow_layout[-1] = {"right": 1.0, "y": calc_x0(y, self.height)}
        elif base_layout_key == "right_mid":
            arrow_layout[-1] = {"left": 0.0, "y": calc_x0(y, self.height)}
        return arrow_layout

    def update_content_size(self, instance, value):
        self.content_size = self.content.size

    def adjust_position(self):
        if self.limit_to is not None and not self._temporarily_ignore_limits:
            if self.limit_to is EventLoop.window:
                lim_x, lim_y = 0, 0
                lim_top, lim_right = self.limit_to.size
            else:
                lim_x = self.limit_to.x
                lim_y = self.limit_to.y
                lim_top = self.limit_to.top
                lim_right = self.limit_to.right

            self._temporarily_ignore_limits = True

            if not (lim_x > self.x and lim_right < self.right):
                self.x = max(lim_x, min(lim_right - self.width, self.x))

            if not (lim_y > self.y and lim_right < self.right):
                self.y = min(lim_top - self.height, max(lim_y, self.y))

            self._temporarily_ignore_limits = False

    def reposition_inner_widgets(self):
        arrow_image_layout = self._arrow_image_layout
        arrow_image_scatter = self._arrow_image_scatter
        arrow_image_scatter_wrapper = self._arrow_image_scatter_wrapper
        content = self.content

        # Remove the children of the Bubble (BoxLayout) as a first step
        for child in list(self.children):
            super().remove_widget(child)

        if self.canvas is None or content is None:
            return

        # find the layout parameters that define a specific bubble setup
        if self._flex_arrow_layout_params is not None:
            layout_params = self._flex_arrow_layout_params
        else:
            layout_params = Bubble.ARROW_LAYOUTS[self.arrow_pos]
        (bubble_orientation,
         widget_order,
         arrow_size_hint,
         arrow_rotation,
         arrow_pos_hint) = layout_params

        # rotate the arrow, place it at the right pos and setup the size
        # of the widget, so the BoxLayout can do the rest.
        arrow_image_scatter.rotation = arrow_rotation
        arrow_image_scatter_wrapper.size = arrow_image_scatter.bbox[1]
        arrow_image_scatter_wrapper.pos_hint = arrow_pos_hint
        arrow_image_layout.size_hint = arrow_size_hint
        arrow_image_layout.size = arrow_image_scatter.bbox[1]

        # set the orientation of the Bubble (BoxLayout)
        self.orientation = bubble_orientation

        # Add the updated children of the Bubble (BoxLayout) and update
        # properties
        widgets_to_add = [content, arrow_image_layout]

        # Set the arrow_margin, so we can use this property for proper sizing
        # of the Bubble Widget.
        # Determine whether to add the arrow_image_layout to the
        # Bubble (BoxLayout) or not.
        arrow_margin_x, arrow_margin_y = (0, 0)
        if self.show_arrow:
            if bubble_orientation[0] == "h":
                arrow_margin_x = arrow_image_layout.width
            elif bubble_orientation[0] == "v":
                arrow_margin_y = arrow_image_layout.height
        else:
            widgets_to_add.pop(1)

        for widget in widgets_to_add[::widget_order]:
            super().add_widget(widget)

        self.arrow_margin = (arrow_margin_x, arrow_margin_y)



"""
Color Wheel and Color Picker Tests
==================================
"""

from kivy.tests.common import GraphicUnitTest, UTMotionEvent
from kivy.uix.colorpicker import ColorWheel, ColorPicker


class ColorWheelTest(GraphicUnitTest):
    def test_render(self):
        color_wheel = ColorWheel()
        self.render(color_wheel)

    def test_clicks(self):
        color_wheel = ColorWheel()

        # ColorPicker has a stated default colour (opaque white).
        # ColorWheel has a different default Color (transparent black).
        self.assertEqual(color_wheel.color, [0, 0, 0, 0])

        # Click on corner of widget
        pos = (color_wheel.pos[0], color_wheel.pos[1])
        touch = UTMotionEvent(
            "unittest",
            1,
            {
                "x": pos[0],
                "y": pos[1],
            },
        )
        touch.grab_current = color_wheel
        touch.pos = pos

        color_wheel.on_touch_down(touch)
        color_wheel.on_touch_up(touch)

        # Too far from the center. No effect.
        self.assertEqual(color_wheel.color, [0, 0, 0, 0])

        pos = (
            color_wheel.pos[0] + color_wheel.size[0] / 2,
            color_wheel.pos[1] + color_wheel.size[1] / 4,
        )
        # Click in middle, half-the-radius up.
        touch = UTMotionEvent(
            "unittest",
            1,
            {"x": pos[0], "y": pos[1]},
        )
        touch.grab_current = color_wheel
        touch.pos = pos

        color_wheel.on_touch_down(touch)
        color_wheel.on_touch_up(touch)

        self.assertEqual(color_wheel.color, [0.75, 0.5, 1, 1])


class ColorPickerTest(GraphicUnitTest):
    def test_render(self):
        color_picker = ColorPicker()
        self.render(color_picker)

    def test_set_colour(self):
        color_picker = ColorPicker()
        # ColorPicker has a stated default colour (opaque white).
        # ColorWheel has a different default Color (transparent black).
        self.assertEqual(color_picker.color, [1, 1, 1, 1])

        # Set without alpha
        color_picker.set_color((0.5, 0.6, 0.7))
        self.assertEqual(color_picker.color, [0.5, 0.6, 0.7, 1])

        # Set with alpha
        color_picker.set_color((0.5, 0.6, 0.7, 0.8))
        self.assertEqual(color_picker.color, [0.5, 0.6, 0.7, 0.8])



from kivy.graphics.fbo cimport Fbo
from kivy.graphics.context_instructions cimport Translate, Scale
from kivy.graphics.vertex_instructions cimport VertexInstruction
from kivy.graphics.instructions cimport InstructionGroup


cdef class BoxShadow(InstructionGroup):

    cdef bint _inset
    cdef float _blur_radius
    cdef tuple _pos
    cdef tuple _size
    cdef tuple _offset
    cdef tuple _border_radius
    cdef tuple _spread_radius
    cdef VertexInstruction _fbo_rect
    cdef VertexInstruction _texture_container
    cdef Scale _fbo_scale
    cdef Translate _fbo_translate
    cdef Fbo _fbo

    cdef void _init_texture(self)
    cdef void _update_canvas(self)
    cdef void _update_fbo(self)
    cdef void _update_shadow(self)
    cdef tuple _adjusted_pos(self)
    cdef tuple _adjusted_size(self)
    cdef object _bounded_value(self, object value, min_value=?, max_value=?)
    cdef bint _check_bool(self, object value)
    cdef float _check_float(self, str property_name, object value, str iter_text=?)
    cdef tuple _check_iter(self, str property_name, object value, int components=?)
    



'''
Bubble unit tests
=================
author: Anthony Zimmermann (Anthony.Zimmermann@protonmail.com)
'''

import pytest
from kivy.tests.common import GraphicUnitTest

from kivy.base import EventLoop

from kivy.uix.bubble import Bubble
from kivy.uix.bubble import BubbleContent
from kivy.uix.bubble import BubbleButton


class _TestBubbleButton(BubbleButton):

    def __init__(self, button_size=(None, None), *args, **kwargs):
        super().__init__(*args, **kwargs)

        size_x, size_y = button_size
        if size_x is not None:
            self.size_hint_x = None
            self.width = size_x
        if size_y is not None:
            self.size_hint_y = None
            self.height = size_y


class _TestBubbleContent(BubbleContent):

    def update_size(self, instance, value):
        self.size = self.minimum_size

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.bind(minimum_size=self.update_size)


class _TestBubble(Bubble):

    @property
    def arrow_length(self):
        return self._arrow_image.height

    @property
    def arrow_width(self):
        return self._arrow_image.width

    @property
    def arrow_rotation(self):
        return self._arrow_image_scatter.rotation

    @property
    def arrow_layout_pos(self):
        return self._arrow_image_layout.pos

    @property
    def arrow_layout_size(self):
        return self._arrow_image_layout.size

    @property
    def arrow_center_pos_within_arrow_layout(self):
        x = self._arrow_image_scatter_wrapper.center_x
        y = self._arrow_image_scatter_wrapper.center_y
        return x, y


# (bubble_width, button_height, arrow_pos)
bubble_layout_with_predefined_arrow_pos_test_params = [
    (158.9, 34.3, "bottom_left"),   # noqa: E201,E241
    (651.4, 26.1, "bottom_mid"),    # noqa: E201,E241
    (  6.5, 44.7, "bottom_right"),  # noqa: E201,E241
    (754.6, 50.6, "top_left"),      # noqa: E201,E241
    (957.8, 74.1, "top_mid"),       # noqa: E201,E241
    (852.1, 33.1, "top_right"),     # noqa: E201,E241
    (472.9, 45.1, "left_top"),      # noqa: E201,E241
    (578.3, 52.7, "left_mid"),      # noqa: E201,E241
    (687.8, 17.7, "left_bottom"),   # noqa: E201,E241
    (313.7,  8.6, "right_top"),     # noqa: E201,E241
    (194.3, 46.4, "right_mid"),     # noqa: E201,E241
    ( 59.3, 29.7, "right_bottom"),  # noqa: E201,E241
]

bubble_layout_with_flex_arrow_pos_test_params = [
    (101.3, 346.0,   0.0,  73.6, "l"),  # noqa: E201,E241
    (489.0, 535.1,   0.0, 442.1, "l"),  # noqa: E201,E241
    (390.9, 728.1,   0.0, 114.3, "l"),  # noqa: E201,E241
    (224.5, 675.5,   0.0, 560.6, "l"),  # noqa: E201,E241
    (264.9, 677.3,   0.0,   8.3, "l"),  # noqa: E201,E241
    (544.6, 126.0,   0.0, 120.9, "l"),  # noqa: E201,E241
    (290.9, 962.6,   0.0, 275.5, "l"),  # noqa: E201,E241
    (358.4, 514.4,   0.0, 427.1, "l"),  # noqa: E201,E241
    (604.3, 648.2,   0.0, 226.1, "l"),  # noqa: E201,E241
    (287.4, 875.6,   0.0, 446.5, "l"),  # noqa: E201,E241
    (755.7, 103.5, 444.6,   0.0, "b"),  # noqa: E201,E241
    (307.9, 471.7,  80.9,   0.0, "b"),  # noqa: E201,E241
    (849.9, 194.8, 652.7,   0.0, "b"),  # noqa: E201,E241
    (975.7, 691.0, 120.9,   0.0, "b"),  # noqa: E201,E241
    (539.1, 903.3, 530.6,   0.0, "b"),  # noqa: E201,E241
    ( 37.5, 727.5,  37.0,   0.0, "b"),  # noqa: E201,E241
    (856.5, 779.0, 565.5,   0.0, "b"),  # noqa: E201,E241
    (536.7, 228.3,  48.4,   0.0, "b"),  # noqa: E201,E241
    (170.9, 870.4, 127.6,   0.0, "b"),  # noqa: E201,E241
    (955.7, 530.6, 526.0,   0.0, "b"),  # noqa: E201,E241
    (878.1, 690.4, 878.1,  18.8, "r"),  # noqa: E201,E241
    (771.6, 365.2, 771.6,  31.1, "r"),  # noqa: E201,E241
    (679.7, 305.4, 679.7, 259.6, "r"),  # noqa: E201,E241
    (700.2, 614.6, 700.2, 105.4, "r"),  # noqa: E201,E241
    (444.1, 864.5, 444.1, 152.3, "r"),  # noqa: E201,E241
    (189.0, 790.4, 189.0, 602.9, "r"),  # noqa: E201,E241
    (376.0, 993.9, 376.0, 486.4, "r"),  # noqa: E201,E241
    (518.5, 338.5, 518.5, 194.6, "r"),  # noqa: E201,E241
    (982.1, 666.1, 982.1, 282.5, "r"),  # noqa: E201,E241
    (926.4, 565.1, 926.4, 187.3, "r"),  # noqa: E201,E241
    (375.2, 746.6,  36.2, 746.6, "t"),  # noqa: E201,E241
    (448.9, 228.5, 297.4, 228.5, "t"),  # noqa: E201,E241
    (792.3, 593.5, 746.2, 593.5, "t"),  # noqa: E201,E241
    (856.1,  89.7,  23.1,  89.7, "t"),  # noqa: E201,E241
    (721.3, 319.0, 356.5, 319.0, "t"),  # noqa: E201,E241
    (127.7, 355.7,  69.3, 355.7, "t"),  # noqa: E201,E241
    (412.3, 493.8, 163.2, 493.8, "t"),  # noqa: E201,E241
    ( 40.8, 115.8,  15.8, 115.8, "t"),  # noqa: E201,E241
    (233.9, 148.5, 189.4, 148.5, "t"),  # noqa: E201,E241
    (982.4, 661.5, 105.9, 661.5, "t"),  # noqa: E201,E241
]


class BubbleTest(GraphicUnitTest):

    def move_frames(self, t):
        for i in range(t):
            EventLoop.idle()

    def test_no_content(self):
        bubble = Bubble()
        self.render(bubble)

    def test_add_remove_content(self):
        bubble = Bubble()
        content = BubbleContent()
        bubble.add_widget(content)
        self.render(bubble)

        bubble.remove_widget(content)
        self.render(bubble)

    def test_add_arbitrary_content(self):
        from kivy.uix.gridlayout import GridLayout
        bubble = Bubble()
        content = GridLayout()
        bubble.add_widget(content)
        self.render(bubble)

    def test_add_two_content_widgets_fails(self):
        from kivy.uix.bubble import BubbleException
        bubble = Bubble()
        content_1 = BubbleContent()
        content_2 = BubbleContent()
        bubble.add_widget(content_1)
        with self.assertRaises(BubbleException):
            bubble.add_widget(content_2)

    def test_add_content_with_buttons(self):
        bubble = Bubble()

        content = BubbleContent()
        content.add_widget(BubbleButton(text="Option A"))
        content.add_widget(BubbleButton(text="Option B"))

        bubble.add_widget(content)
        self.render(bubble)

    def assertSequenceAlmostEqual(self, seq1, seq2, delta=None):
        assert len(seq1) == len(seq2)
        for a, b in zip(seq1, seq2):
            self.assertAlmostEqual(a, b, delta=delta)

    def assertTestBubbleLayoutWithPredefinedArrowPos(self, bubble):
        arrow_length = bubble.arrow_length
        arrow_width = bubble.arrow_width
        bubble_width = bubble.test_bubble_width
        button_height = bubble.test_button_height

        # assert content size
        expected_content_size = {
            "bottom_left": (bubble_width, button_height),
            "bottom_mid": (bubble_width, button_height),
            "bottom_right": (bubble_width, button_height),
            "top_left": (bubble_width, button_height),
            "top_mid": (bubble_width, button_height),
            "top_right": (bubble_width, button_height),
            "left_top": (bubble_width - arrow_length, button_height),
            "left_mid": (bubble_width - arrow_length, button_height),
            "left_bottom": (bubble_width - arrow_length, button_height),
            "right_top": (bubble_width - arrow_length, button_height),
            "right_mid": (bubble_width - arrow_length, button_height),
            "right_bottom": (bubble_width - arrow_length, button_height),
        }[bubble.arrow_pos]
        self.assertSequenceAlmostEqual(
            bubble.content.size,
            expected_content_size,
        )

        # assert arrow layout size
        expected_arrow_layout_size = {
            "bottom_left": (bubble_width, arrow_length),
            "bottom_mid": (bubble_width, arrow_length),
            "bottom_right": (bubble_width, arrow_length),
            "top_left": (bubble_width, arrow_length),
            "top_mid": (bubble_width, arrow_length),
            "top_right": (bubble_width, arrow_length),
            "left_top": (arrow_length, button_height),
            "left_mid": (arrow_length, button_height),
            "left_bottom": (arrow_length, button_height),
            "right_top": (arrow_length, button_height),
            "right_mid": (arrow_length, button_height),
            "right_bottom": (arrow_length, button_height),
        }[bubble.arrow_pos]
        self.assertSequenceAlmostEqual(
            bubble.arrow_layout_size,
            expected_arrow_layout_size,
        )

        # assert content position
        expected_content_position = {
            "bottom_left": (0, arrow_length),
            "bottom_mid": (0, arrow_length),
            "bottom_right": (0, arrow_length),
            "top_left": (0, 0),
            "top_mid": (0, 0),
            "top_right": (0, 0),
            "left_top": (arrow_length, 0),
            "left_mid": (arrow_length, 0),
            "left_bottom": (arrow_length, 0),
            "right_top": (0, 0),
            "right_mid": (0, 0),
            "right_bottom": (0, 0),
        }[bubble.arrow_pos]
        self.assertSequenceAlmostEqual(
            bubble.content.pos,
            expected_content_position,
        )

        # assert arrow layout position
        expected_arrow_layout_position = {
            "bottom_left": (0, 0),
            "bottom_mid": (0, 0),
            "bottom_right": (0, 0),
            "top_left": (0, button_height),
            "top_mid": (0, button_height),
            "top_right": (0, button_height),
            "left_top": (0, 0),
            "left_mid": (0, 0),
            "left_bottom": (0, 0),
            "right_top": (bubble_width - arrow_length, 0),
            "right_mid": (bubble_width - arrow_length, 0),
            "right_bottom": (bubble_width - arrow_length, 0),
        }[bubble.arrow_pos]
        self.assertSequenceAlmostEqual(
            bubble.arrow_layout_pos,
            expected_arrow_layout_position,
        )

        # assert arrow position within arrow layout
        hal = arrow_length / 2  # hal := half arrow length
        x_offset = 0.05 * bubble_width
        y_offset = 0.05 * button_height
        expected_arrow_center_pos_within_arrow_layout = {
            "bottom_left": (x_offset + arrow_width / 2, hal),
            "bottom_mid": (bubble_width / 2, hal),
            "bottom_right": (bubble_width - arrow_width / 2 - x_offset, hal),
            "top_left": (x_offset + arrow_width / 2, hal),
            "top_mid": (bubble_width / 2, hal),
            "top_right": (bubble_width - arrow_width / 2 - x_offset, hal),
            "left_top": (hal, button_height - arrow_width / 2 - y_offset),
            "left_mid": (hal, button_height / 2),
            "left_bottom": (hal, y_offset + arrow_width / 2),
            "right_top": (hal, button_height - arrow_width / 2 - y_offset),
            "right_mid": (hal, button_height / 2),
            "right_bottom": (hal, y_offset + arrow_width / 2),
        }[bubble.arrow_pos]
        self.assertSequenceAlmostEqual(
            bubble.arrow_center_pos_within_arrow_layout,
            expected_arrow_center_pos_within_arrow_layout,
        )

        # assert arrow rotation
        expected_arrow_rotation = {
            "bottom_left": 0,
            "bottom_mid": 0,
            "bottom_right": 0,
            "top_left": 180,
            "top_mid": 180,
            "top_right": 180,
            "left_top": 270,
            "left_mid": 270,
            "left_bottom": 270,
            "right_top": 90,
            "right_mid": 90,
            "right_bottom": 90,
        }[bubble.arrow_pos]
        self.assertAlmostEqual(
            bubble.arrow_rotation,
            expected_arrow_rotation,
        )

    def test_bubble_layout_with_predefined_arrow_pos(self):

        for params in bubble_layout_with_predefined_arrow_pos_test_params:
            bubble_width, button_height, arrow_pos = params

            with self.subTest():
                print(
                    "(bubble_width={}, button_height={}, arrow_pos={})".format(
                        *params
                    )
                )
                bubble = _TestBubble(arrow_pos=arrow_pos)
                bubble.size_hint = (None, None)
                bubble.test_bubble_width = bubble_width
                bubble.test_button_height = button_height

                def update_bubble_size(instance, value):
                    w = bubble_width
                    h = bubble.content_height + bubble.arrow_margin_y
                    bubble.size = (w, h)
                bubble.bind(
                    content_size=update_bubble_size,
                    arrow_margin=update_bubble_size,
                )

                content = _TestBubbleContent()
                for i in range(3):
                    content.add_widget(
                        _TestBubbleButton(
                            button_size=(None, button_height),
                            text="Option {}".format(i)
                        )
                    )

                bubble.add_widget(content)
                self.render(bubble)

                self.assertTestBubbleLayoutWithPredefinedArrowPos(bubble)

    def test_bubble_layout_without_arrow(self):
        bubble_width = 200
        button_height = 30

        bubble = _TestBubble(show_arrow=False)
        bubble.size_hint = (None, None)

        def update_bubble_size(instance, value):
            w = bubble_width
            h = bubble.content_height
            bubble.size = (w, h)
        bubble.bind(content_size=update_bubble_size)

        content = _TestBubbleContent(orientation="vertical")
        for i in range(7):
            content.add_widget(
                _TestBubbleButton(
                    button_size=(None, button_height),
                    text="Option_{}".format(i)
                )
            )

        bubble.add_widget(content)
        self.render(bubble)

        # assert content size
        self.assertSequenceAlmostEqual(
            bubble.content.size,
            (bubble_width, 7 * button_height),
        )

        # assert content position
        self.assertSequenceAlmostEqual(
            bubble.content.pos,
            (0, 0),
        )

    def test_bubble_layout_with_flex_arrow_pos(self):
        for params in bubble_layout_with_flex_arrow_pos_test_params:
            bubble_size = params[:2]
            flex_arrow_pos = params[2:4]
            arrow_side = params[4]
            with self.subTest():
                print("(w={}, h={}, x={}, y={}, side={})".format(*params))

                bubble = _TestBubble()
                bubble.size_hint = (None, None)
                bubble.size = bubble_size
                bubble.flex_arrow_pos = flex_arrow_pos

                content = _TestBubbleContent(orientation="vertical")
                content.size_hint = (1, 1)

                button = _TestBubbleButton(
                    button_size=(None, None),
                    text="Option",
                )
                button.size_hint_y = 1

                content.add_widget(button)
                bubble.add_widget(content)

                self.render(bubble)

                haw = bubble.arrow_width / 2  # half arrow_width
                if arrow_side in ["l", "r"]:
                    self.assertSequenceAlmostEqual(
                        bubble.arrow_center_pos_within_arrow_layout,
                        (haw, flex_arrow_pos[1]),
                        delta=haw,
                    )
                elif arrow_side in ["b", "t"]:
                    self.assertSequenceAlmostEqual(
                        bubble.arrow_center_pos_within_arrow_layout,
                        (flex_arrow_pos[0], haw),
                        delta=haw,
                    )



import os
from tempfile import TemporaryDirectory

from unittest import mock


def test_garden_import_module():
    with TemporaryDirectory() as tmpdir:
        fake_garden_dir = os.path.join(tmpdir, "garden")

        with mock.patch("kivy.garden.garden_system_dir", fake_garden_dir):
            # Add a fake garden module to the temporary directory
            dummy_flower_dir = os.path.join(
                fake_garden_dir, "garden.dummyflower"
            )
            os.makedirs(dummy_flower_dir, exist_ok=True)

            with open(os.path.join(dummy_flower_dir, "__init__.py"), "w") as f:
                f.write("dummy_var = 1")

            from kivy.garden.dummyflower import dummy_var

            assert dummy_var == 1
